#!/usr/bin/perl


$tx = shift || 0;
$ty = shift || 0;
$tz = shift || 0;

$seed = shift || 0;

# Build a tree branch recursively

$p2 = 0.7; $p3 = 0.3; 
$rx1min = -10; $rx1max = 0;
$ry1min = -10;  $ry1max = 10;
$rz1min = -2; $rz1max = 2;

$rx2min = -10; $rx2max = 0;
$ry2min = -10;  $ry2max = 10;
$rz2min = -2; $rz2max = 2;

$bx2 = $bz2 = 0; $by2 = 45;

$bpct = 0.4;
$minbi = 2;  $bia = 2; $bim = 1;

$minlen = 0.01;
$minth = 0.01;
$needleth = 0.25;

$pi = atan2(0,-1);

sub len {
    my($x,$y,$z) = @_;
    return sqrt($x*$x + $y*$y+ $z*$z);
}

sub sign {
    $_[0] < 0 ? -1 : 1;
}

sub rad2deg {
    # dunno why but POV uses degrees...

    return (180 * $_[0]) / $pi ;
}
sub deg2rad {
    # dunno why but POV uses degrees...

    return ($_[0] * $pi) / 180 ;
}


# rotate a vector the given x,y,z in degrees (in deference to POV) 
sub rotate {
    my($x, $y, $z, $rx, $ry, $rz) = @_;
    my($nx, $ny, $nz);
    my(@xrot, @yrot, @zrot,@a, @b);
    my($i, $j);
    $rx = deg2rad($rx);
    $ry = deg2rad($ry);
    $rz = deg2rad($rz);


    # rotate around x
    if ( $rx ) {
	$ny = $y * cos($rx) - $z * sin($rx);
	$nz = $y * sin($rx) + $z * cos($rx);
	$y = $ny; $z = $nz;
    }

    # rotate around y

    if ( $ry ) {
	$nx = $z * sin($ry) + $x * cos($ry);
	$nz = $z * cos($ry) - $x * sin($ry);
	$x = $nx; $z = $nz;
    }
    # rotate around z

    if ($rz) {
	$nx = $x * cos($rz) - $y * sin($rz);
	$ny = $x * sin($rz) + $y * cos($rz);
	$x = $nx; $y = $ny;
    }

    return ($x, $y, $z);
}

sub branch {
    # x, y, z of the base and the orientation vector (the z vector
    # should be moved over it ).  Then thickness, reduction pct, and length
    my($bx, $by, $bz, $ox, $oy, $oz, $rx, $ry, $rz, $th, $red, $len) = @_;
    my ($rtx, $rty, $rtz);  # Point were this branch ended
    my($l);		# Length of the orientation vector
    my($ozp);	# the orientntation vector's z component after we rotate
                # into the x z plane
    my($t1, $t2); # rotation angles for realigment
    my($capr) = $th * $red;	# The cap radius

    $l = len($ox, $oy, $oz);

    #make the orientation vector a unit vector
    if ( $l != 1 ) {
	$ox = $ox / $l;
	$oy = $oy / $l;
	$oz = $oz / $l;
    }

    $t1 = $t2 = 0;
    unless ( $ox == $oy && $ox == 0 ) {
	# we need to reorient things (rotate the ov into the xz plane

#	printf STDERR "oy $oy oz $oz\n";
	$t1 = ($oy == 0) ? 0 : atan2($oy,$oz);
	$ozp = $oy * sin($t1) + $oz * cos($t1);
	unless ( $ox == 0 ) {

	    # now we need to find a second rotation about y.  Because
	    #we're starting with a vector pointed at z, we need to swap
	    #x and z (i.e. consider z the x coordinate and x the y for
	    #2-d trig to calculate rotation angle).

            # atan2 seems to be broken for 0/1 - doh!

#	    print STDERR "$ozp $ox\n";
	   
	    $t2 = ($ox == 0 ) ?
		($ozp / abs($ozp) * $pi ) :
		    atan2($ox,$ozp);
	}
    }

    # this -1 may come from the left handed coordinates.  It works, though.

    $t1 = - rad2deg($t1);
    $t2 = rad2deg($t2);

    # (We're really doing the inverse transform - getting <0, 0, 1> to line
    # up with <$ox, $oy, $oz> , so we actually apply the rotations in the
    # reverse order that we calculated them.)
    print <<END;
cone {
    <0,0,0>, $th , <0,0,$len>, $capr
END
    print "\trotate < 0, $t2, 0>\n" if $t2 != 0;
    print "\trotate < $t1, 0, 0>\n" if $t1 != 0;
    print "\trotate < $rx, 0, 0 >\n" if $rx != 0;
    print "\trotate < 0, $ry, 0 >\n" if $ry != 0;
    print "\trotate < 0, 0, $rz >\n" if $rz != 0;
    print "\ttranslate < $bx, $by, $bz>\n"
	unless $bx == $by && $by == $bz && $bx == 0;
    print <<END;
    texture { TreeTexture } 
}

sphere {
    <0,0,$len>, $capr
END
    print "\trotate < 0, $t2, 0>\n" if $t2 != 0;
    print "\trotate < $t1, 0, 0>\n" if $t1 != 0;
    print "\trotate < $rx, 0, 0 >\n" if $rx != 0;
    print "\trotate < 0, $ry, 0 >\n" if $ry != 0;
    print "\trotate < 0, 0, $rz >\n" if $rz != 0;
    print "\ttranslate < $bx, $by, $bz>\n"
	unless $bx == $by && $by == $bz && $bx == 0;
    print <<END;
    texture { TreeTexture } 
}
END
    if ( $th < $needleth ) {
	print <<END;
#declare n = 2;
#declare i = 0;
#declare ring=2;
#while ( i < n ) 
cone {
	<0,0,$len/n*i>,$th - (i/n * ($th - $capr)), <0,0,$len/n*(i+n)>, $th*3
	open
#declare j = 0;
#while (j < ring) 
	texture {
	    NeedleTexture
	    scale <$th*3, 1, 1>
	    rotate < 0, 0, 180/ring * j + i * 22.5 >
	}
#declare j = j+1;
#end
END
        print "\trotate < 0, $t2, 0>\n" if $t2 != 0;
	print "\trotate < $t1, 0, 0>\n" if $t1 != 0;
	print "\trotate < $rx, 0, 0 >\n" if $rx != 0;
	print "\trotate < 0, $ry, 0 >\n" if $ry != 0;
	print "\trotate < 0, 0, $rz >\n" if $rz != 0;
	print "\ttranslate < $bx, $by, $bz>\n"
	    unless $bx == $by && $by == $bz && $bx == 0;
	print <<END;
}
#declare i = i + 1;
#end
END
    }
    ($ox, $oy, $oz ) = rotate($ox, $oy, $oz, $rx, $ry, $rz ) ;
    ($rtx , $rty, $rtz) = ($bx + $ox * $len, $by + $oy * $len, $bz + $oz * $len);
    return ($rtx, $rty, $rtz, $ox, $oy, $oz);

}

sub tree {
    my($bx,$by,$bz, $ox, $oy, $oz, $th, $red, $len, $d, $b, $bi) = @_;
    my($nox, $noy, $noz);
    my($x, $y, $z);
    my($p);
    my($drx, $dry, $drz);
    my($nb, $nbi);
    my($xdir, $ydir,$zdir);

    return if ( $len < $minlen || $d <= 0 || $th < $minth);

    if ( $b > 0 ) {
	# Because I want to always think of the rotations as though
	# they are performed on the z axis, we have some sign games to
	# play.

	($drx, $dry, $drz ) =
	    (sign($oz) * ($rx1min + rand(abs($rx1max-$rx1min))),
	     sign($oz) * ($ry1min + rand(abs($ry1max-$ry1min))),
	     ($rz1min + rand(abs($rz1max-$rz1min))));

	($x, $y, $z, $nox, $noy, $noz) =
	    branch($bx, $by, $bz, $ox, $oy, $oz, $drx,$dry, $drz, $th, $red, $len);
	tree($x, $y, $z, $nox, $noy, $noz, $th*$red, $red, $len, $d-1, $b-1, $bi);
    } else {
	$xdir = ( rand(1) < 0.5 ) ? 1 : -1;
	$ydir = ( rand(1) < 0.5 ) ? 1 : -1;
	$zdir = ( rand(1) < 0.5 ) ? 1 : -1;
	
	($drx, $dry, $drz ) =
	    (sign($oz) * ($rx2min + rand(abs($rx2max-$rx2min))),
	     sign($oz) * ($ry2min + rand(abs($ry2max-$ry2min))),
	     ($rz2min + rand(abs($rz2max-$rz2min))));

	$nb = int($bi - $bpct * $bi + rand(2*$bpct*$bi));
	$nbi = ($bi - $bia ) / $bim;

	$nbi = $minbi if $nbi < $minbi;
    
	($x, $y, $z, $nox, $noy, $noz) =
	    branch($bx, $by, $bz, $ox, $oy, $oz, $drx+$xdir*$bx2, $ydir*$by2 + $dry, $zdir*$bz2 + $drz, $th, $red, $len);
	tree($x, $y, $z, $nox, $noy, $noz, $th*$red, $red, $len, $d-1,$nb, $nbi);
	($drx, $dry, $drz ) =
	    ( rand(2*$rx2)-$rx2,rand(2*$ry2)-$ry2,rand(2*$rz2)-$rz2);
    
	($x, $y, $z, $nox, $noy, $noz) =
	    branch($bx, $by, $bz, $ox, $oy, $oz, 0, 0, 0, $th, $red, $len);
	$nb = int($bi - $bpct * $bi + rand(2*$bpct*$bi));
	print STDERR "$bx, $by, $bz\n" if $nb > $d-1;
	tree($x, $y, $z, $nox, $noy, $noz, $th*$red, $red, $len, $d-1, $nb, $nbi);
	$p = rand(1);
	if ( $p > $p2 ) {
	    # 3 way branch!
	    ($drx, $dry, $drz ) =
		(sign($oz) * ($rx2min+rand(abs($rx2max-$rx2min))),
		 sign($oz) * ($ry2min+rand(abs($ry2max-$ry2min))),
		 ($rz2min+rand(abs($rz2max-$rz2min))));

	    $nb = int($bi - $bpct * $bi + rand(2*$bpct*$bi));
	    ($x, $y, $z, $nox, $noy, $noz) =
		branch($bx, $by, $bz, $ox, $oy, $oz, $drx-$xdir*$bx2, -$ydir*$by2 + $dry, -$zdir*$bz2 + $drz, $th, $red, $len);
	    tree($x, $y, $z, $nox, $noy, $noz, $th*$red, $red, $len, $d-1,$nb, $nbi);
	}
	    
    }
}

$max = 72; $inc = 3;

srand($seed ) if ( $seed) ;

for ( $i = 0; $i < $max-12; $i+= $inc ) {
    $th = 1-$i/$max;
    $len = 1.25 * (1-$i/$max);
    if ( $i % 2 == 0 ) {
	print "union {\n";
	tree($tx,$ty+$i+8,$tz,0,0,1,$th, 0.9, $len, 16, 5, 5);
	print "}\n";
	print "union {\n";
	tree($tx,$ty+$i+8,$tz,0,0,-1,$th, 0.9, $len, 16, 5, 5);
	print "}\n";
	print "union {\n";
	tree($tx,$ty+$i+8,$tz,1,0,0,$th, 0.9, $len, 16, 5, 5);
	print "}\n";
	print "union {\n";
	tree($tx,$ty+$i+8,$tz,-1,0,0,$th, 0.9, $len, 16, 5, 5);
	print "}\n";
    }
    else {
	print "union {\n";
	tree($tx,$ty+$i+8,$tz,1,0,1,$th, 0.9, $len, 16, 5, 5);
	print "}\n";
	print "union {\n";
	tree($tx,$ty+$i+8,$tz,1,0,-1,$th, 0.9, $len, 16, 5, 5);
	print "}\n";
	print "union {\n";
	tree($tx,$ty+$i+8,$tz,-1,0,1,$th, 0.9, $len, 16, 5, 5);
	print "}\n";
	print "union {\n";
	tree($tx,$ty+$i+8,$tz,-1,0,-1,$th, 0.9, $len, 16, 5, 5);
	print "}\n";
    }
}

$rx1min = -1; $rx1max = 1;
$ry1min = -1;  $ry1max = 1;
$rz1min = -1; $rz1max = 1;

tree($tx,$ty+66,$tz,0,1,0,0.3, 0.9, 0.25, 20, 200, 100);
print <<END;
cone {
    <0,0,0>, 1.5, <0,66,0>, 0.3
	texture { TreeTexture }
    translate <$tx, $ty, $tz>
}
END

